@@ -10,12 +10,17 @@
 #endif
 #include "ctypes.h"
 
+
+#define CTYPES_CAPSULE_WCHAR_T "_ctypes/cfield.c wchar_t buffer from unicode"
+CTYPES_CAPSULE_INSTANTIATE_DESTRUCTOR(CTYPES_CAPSULE_WCHAR_T)
+
+
 /******************************************************************/
 /*
-  CField_Type
+  PyCField_Type
 */
 static PyObject *
-CField_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
+PyCField_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
     CFieldObject *obj;
     obj = (CFieldObject *)type->tp_alloc(type, 0);
@@ -35,7 +40,7 @@ CField_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
  * prev_desc points to the type of the previous bitfield, if any.
  */
 PyObject *
-CField_FromDesc(PyObject *desc, Py_ssize_t index,
+PyCField_FromDesc(PyObject *desc, Py_ssize_t index,
                 Py_ssize_t *pfield_size, int bitsize, int *pbitofs,
                 Py_ssize_t *psize, Py_ssize_t *poffset, Py_ssize_t *palign,
                 int pack, int big_endian)
@@ -52,7 +57,7 @@ CField_FromDesc(PyObject *desc, Py_ssize_t index,
 #define CONT_BITFIELD 2
 #define EXPAND_BITFIELD 3
 
-    self = (CFieldObject *)PyObject_CallObject((PyObject *)&CField_Type,
+    self = (CFieldObject *)PyObject_CallObject((PyObject *)&PyCField_Type,
                                                NULL);
     if (self == NULL)
         return NULL;
@@ -102,7 +107,7 @@ CField_FromDesc(PyObject *desc, Py_ssize_t index,
     /*  Field descriptors for 'c_char * n' are be scpecial cased to
         return a Python string instead of an Array object instance...
     */
-    if (ArrayTypeObject_Check(proto)) {
+    if (PyCArrayTypeObject_Check(proto)) {
         StgDictObject *adict = PyType_stgdict(proto);
         StgDictObject *idict;
         if (adict && adict->proto) {
@@ -113,14 +118,14 @@ CField_FromDesc(PyObject *desc, Py_ssize_t index,
                 Py_DECREF(self);
                 return NULL;
             }
-            if (idict->getfunc == getentry("c")->getfunc) {
-                struct fielddesc *fd = getentry("s");
+            if (idict->getfunc == _ctypes_get_fielddesc("c")->getfunc) {
+                struct fielddesc *fd = _ctypes_get_fielddesc("s");
                 getfunc = fd->getfunc;
                 setfunc = fd->setfunc;
             }
 #ifdef CTYPES_UNICODE
-            if (idict->getfunc == getentry("u")->getfunc) {
-                struct fielddesc *fd = getentry("U");
+            if (idict->getfunc == _ctypes_get_fielddesc("u")->getfunc) {
+                struct fielddesc *fd = _ctypes_get_fielddesc("U");
                 getfunc = fd->getfunc;
                 setfunc = fd->setfunc;
             }
@@ -194,7 +199,7 @@ CField_FromDesc(PyObject *desc, Py_ssize_t index,
 }
 
 static int
-CField_set(CFieldObject *self, PyObject *inst, PyObject *value)
+PyCField_set(CFieldObject *self, PyObject *inst, PyObject *value)
 {
     CDataObject *dst;
     char *ptr;
@@ -206,12 +211,12 @@ CField_set(CFieldObject *self, PyObject *inst, PyObject *value)
                         "can't delete attribute");
         return -1;
     }
-    return CData_set(inst, self->proto, self->setfunc, value,
+    return PyCData_set(inst, self->proto, self->setfunc, value,
                      self->index, self->size, ptr);
 }
 
 static PyObject *
-CField_get(CFieldObject *self, PyObject *inst, PyTypeObject *type)
+PyCField_get(CFieldObject *self, PyObject *inst, PyTypeObject *type)
 {
     CDataObject *src;
     if (inst == NULL) {
@@ -220,51 +225,51 @@ CField_get(CFieldObject *self, PyObject *inst, PyTypeObject *type)
     }
     assert(CDataObject_Check(inst));
     src = (CDataObject *)inst;
-    return CData_get(self->proto, self->getfunc, inst,
+    return PyCData_get(self->proto, self->getfunc, inst,
                      self->index, self->size, src->b_ptr + self->offset);
 }
 
 static PyObject *
-CField_get_offset(PyObject *self, void *data)
+PyCField_get_offset(PyObject *self, void *data)
 {
     return PyInt_FromSsize_t(((CFieldObject *)self)->offset);
 }
 
 static PyObject *
-CField_get_size(PyObject *self, void *data)
+PyCField_get_size(PyObject *self, void *data)
 {
     return PyInt_FromSsize_t(((CFieldObject *)self)->size);
 }
 
-static PyGetSetDef CField_getset[] = {
-    { "offset", CField_get_offset, NULL, "offset in bytes of this field" },
-    { "size", CField_get_size, NULL, "size in bytes of this field" },
+static PyGetSetDef PyCField_getset[] = {
+    { "offset", PyCField_get_offset, NULL, "offset in bytes of this field" },
+    { "size", PyCField_get_size, NULL, "size in bytes of this field" },
     { NULL, NULL, NULL, NULL },
 };
 
 static int
-CField_traverse(CFieldObject *self, visitproc visit, void *arg)
+PyCField_traverse(CFieldObject *self, visitproc visit, void *arg)
 {
     Py_VISIT(self->proto);
     return 0;
 }
 
 static int
-CField_clear(CFieldObject *self)
+PyCField_clear(CFieldObject *self)
 {
     Py_CLEAR(self->proto);
     return 0;
 }
 
 static void
-CField_dealloc(PyObject *self)
+PyCField_dealloc(PyObject *self)
 {
-    CField_clear((CFieldObject *)self);
+    PyCField_clear((CFieldObject *)self);
     self->ob_type->tp_free((PyObject *)self);
 }
 
 static PyObject *
-CField_repr(CFieldObject *self)
+PyCField_repr(CFieldObject *self)
 {
     PyObject *result;
     Py_ssize_t bits = self->size >> 16;
@@ -292,17 +297,17 @@ CField_repr(CFieldObject *self)
     return result;
 }
 
-PyTypeObject CField_Type = {
+PyTypeObject PyCField_Type = {
     PyVarObject_HEAD_INIT(NULL, 0)
     "_ctypes.CField",                                   /* tp_name */
     sizeof(CFieldObject),                       /* tp_basicsize */
     0,                                          /* tp_itemsize */
-    CField_dealloc,                             /* tp_dealloc */
+    PyCField_dealloc,                                   /* tp_dealloc */
     0,                                          /* tp_print */
     0,                                          /* tp_getattr */
     0,                                          /* tp_setattr */
     0,                                          /* tp_compare */
-    (reprfunc)CField_repr,                      /* tp_repr */
+    (reprfunc)PyCField_repr,                            /* tp_repr */
     0,                                          /* tp_as_number */
     0,                                          /* tp_as_sequence */
     0,                                          /* tp_as_mapping */
@@ -314,23 +319,23 @@ PyTypeObject CField_Type = {
     0,                                          /* tp_as_buffer */
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC, /* tp_flags */
     "Structure/Union member",                   /* tp_doc */
-    (traverseproc)CField_traverse,              /* tp_traverse */
-    (inquiry)CField_clear,                      /* tp_clear */
+    (traverseproc)PyCField_traverse,                    /* tp_traverse */
+    (inquiry)PyCField_clear,                            /* tp_clear */
     0,                                          /* tp_richcompare */
     0,                                          /* tp_weaklistoffset */
     0,                                          /* tp_iter */
     0,                                          /* tp_iternext */
     0,                                          /* tp_methods */
     0,                                          /* tp_members */
-    CField_getset,                              /* tp_getset */
+    PyCField_getset,                                    /* tp_getset */
     0,                                          /* tp_base */
     0,                                          /* tp_dict */
-    (descrgetfunc)CField_get,                   /* tp_descr_get */
-    (descrsetfunc)CField_set,                   /* tp_descr_set */
+    (descrgetfunc)PyCField_get,                 /* tp_descr_get */
+    (descrsetfunc)PyCField_set,                 /* tp_descr_set */
     0,                                          /* tp_dictoffset */
     0,                                          /* tp_init */
     0,                                          /* tp_alloc */
-    CField_new,                                 /* tp_new */
+    PyCField_new,                               /* tp_new */
     0,                                          /* tp_free */
 };
 
@@ -1191,8 +1196,8 @@ u_set(void *ptr, PyObject *value, Py_ssize_t size)
 
     if (PyString_Check(value)) {
         value = PyUnicode_FromEncodedObject(value,
-                                            conversion_mode_encoding,
-                                            conversion_mode_errors);
+                                            _ctypes_conversion_encoding,
+                                            _ctypes_conversion_errors);
         if (!value)
             return NULL;
     } else if (!PyUnicode_Check(value)) {
@@ -1266,8 +1271,8 @@ U_set(void *ptr, PyObject *value, Py_ssize_t length)
 
     if (PyString_Check(value)) {
         value = PyUnicode_FromEncodedObject(value,
-                                            conversion_mode_encoding,
-                                            conversion_mode_errors);
+                                            _ctypes_conversion_encoding,
+                                            _ctypes_conversion_errors);
         if (!value)
             return NULL;
     } else if (!PyUnicode_Check(value)) {
@@ -1364,8 +1369,8 @@ z_set(void *ptr, PyObject *value, Py_ssize_t size)
         return value;
     } else if (PyUnicode_Check(value)) {
         PyObject *str = PyUnicode_AsEncodedString(value,
-                                                  conversion_mode_encoding,
-                                                  conversion_mode_errors);
+                                                  _ctypes_conversion_encoding,
+                                                  _ctypes_conversion_errors);
         if (str == NULL)
             return NULL;
         *(char **)ptr = PyString_AS_STRING(str);
@@ -1415,8 +1420,8 @@ Z_set(void *ptr, PyObject *value, Py_ssize_t size)
     }
     if (PyString_Check(value)) {
         value = PyUnicode_FromEncodedObject(value,
-                                            conversion_mode_encoding,
-                                            conversion_mode_errors);
+                                            _ctypes_conversion_encoding,
+                                            _ctypes_conversion_errors);
         if (!value)
             return NULL;
     } else if (PyInt_Check(value) || PyLong_Check(value)) {
@@ -1457,7 +1462,7 @@ Z_set(void *ptr, PyObject *value, Py_ssize_t size)
             return PyErr_NoMemory();
         }
         memset(buffer, 0, size);
-        keep = PyCObject_FromVoidPtr(buffer, PyMem_Free);
+        keep = CAPSULE_NEW(buffer, CTYPES_CAPSULE_WCHAR_T);
         if (!keep) {
             Py_DECREF(value);
             PyMem_Free(buffer);
@@ -1509,8 +1514,8 @@ BSTR_set(void *ptr, PyObject *value, Py_ssize_t size)
         value = NULL;
     } else if (PyString_Check(value)) {
         value = PyUnicode_FromEncodedObject(value,
-                                            conversion_mode_encoding,
-                                            conversion_mode_errors);
+                                            _ctypes_conversion_encoding,
+                                            _ctypes_conversion_errors);
         if (!value)
             return NULL;
     } else if (PyUnicode_Check(value)) {
@@ -1670,7 +1675,7 @@ static struct fielddesc formattable[] = {
 */
 
 struct fielddesc *
-getentry(char *fmt)
+_ctypes_get_fielddesc(char *fmt)
 {
     static int initialized = 0;
     struct fielddesc *table = formattable;
@@ -1679,11 +1684,11 @@ getentry(char *fmt)
         initialized = 1;
 #ifdef CTYPES_UNICODE
         if (sizeof(wchar_t) == sizeof(short))
-            getentry("u")->pffi_type = &ffi_type_sshort;
+            _ctypes_get_fielddesc("u")->pffi_type = &ffi_type_sshort;
         else if (sizeof(wchar_t) == sizeof(int))
-            getentry("u")->pffi_type = &ffi_type_sint;
+            _ctypes_get_fielddesc("u")->pffi_type = &ffi_type_sint;
         else if (sizeof(wchar_t) == sizeof(long))
-            getentry("u")->pffi_type = &ffi_type_slong;
+            _ctypes_get_fielddesc("u")->pffi_type = &ffi_type_slong;
 #endif
     }
 
