@@ -511,9 +511,8 @@ audioop_findfit(PyObject *self, PyObject *args)
 
     best_result = result;
     best_j = 0;
-    j = 0;
 
-    for ( j=1; j<=len1-len2; j++) {
+    for (j=1; j<=len1-len2; j++) {
         aj_m1 = (double)cp1[j-1];
         aj_lm1 = (double)cp1[j+len2-1];
 
@@ -597,9 +596,8 @@ audioop_findmax(PyObject *self, PyObject *args)
 
     best_result = result;
     best_j = 0;
-    j = 0;
 
-    for ( j=1; j<=len1-len2; j++) {
+    for (j=1; j<=len1-len2; j++) {
         aj_m1 = (double)cp1[j-1];
         aj_lm1 = (double)cp1[j+len2-1];
 
@@ -1300,7 +1298,7 @@ audioop_ulaw2lin(PyObject *self, PyObject *args)
                            &cp, &len, &size) )
         return 0;
 
-    if (!audioop_check_parameters(len, size))
+    if (!audioop_check_size(size))
         return NULL;
 
     if (len > INT_MAX/size) {
@@ -1369,7 +1367,7 @@ audioop_alaw2lin(PyObject *self, PyObject *args)
                            &cp, &len, &size) )
         return 0;
 
-    if (!audioop_check_parameters(len, size))
+    if (!audioop_check_size(size))
         return NULL;
 
     if (len > INT_MAX/size) {
@@ -1419,7 +1417,6 @@ audioop_lin2adpcm(PyObject *self, PyObject *args)
     if ( state == Py_None ) {
         /* First time, it seems. Set defaults */
         valpred = 0;
-        step = 7;
         index = 0;
     } else if ( !PyArg_ParseTuple(state, "ii", &valpred, &index) )
         return 0;
@@ -1512,14 +1509,13 @@ audioop_adpcm2lin(PyObject *self, PyObject *args)
                            &cp, &len, &size, &state) )
         return 0;
 
-    if (!audioop_check_parameters(len, size))
+    if (!audioop_check_size(size))
         return NULL;
 
     /* Decode state, should have (value, step) */
     if ( state == Py_None ) {
         /* First time, it seems. Set defaults */
         valpred = 0;
-        step = 7;
         index = 0;
     } else if ( !PyArg_ParseTuple(state, "ii", &valpred, &index) )
         return 0;
