@@ -3,6 +3,7 @@
 
 #include "Python.h"
 #include <ctype.h>
+#include <float.h>
 
 static PyObject *int_int(PyIntObject *v);
 
@@ -78,7 +79,8 @@ fill_free_list(void)
 static PyIntObject *small_ints[NSMALLNEGINTS + NSMALLPOSINTS];
 #endif
 #ifdef COUNT_ALLOCS
-int quick_int_allocs, quick_neg_int_allocs;
+Py_ssize_t quick_int_allocs;
+Py_ssize_t quick_neg_int_allocs;
 #endif
 
 PyObject *
@@ -176,7 +178,7 @@ PyInt_AsLong(register PyObject *op)
         {
             Py_DECREF(io);
             PyErr_SetString(PyExc_TypeError,
-                            "nb_int should return int object");
+                        "__int__ method should return an integer");
             return -1;
         }
     }
@@ -234,7 +236,7 @@ PyInt_AsSsize_t(register PyObject *op)
         {
             Py_DECREF(io);
             PyErr_SetString(PyExc_TypeError,
-                            "nb_int should return int object");
+                        "__int__ method should return an integer");
             return -1;
         }
     }
@@ -279,7 +281,7 @@ PyInt_AsUnsignedLongMask(register PyObject *op)
         {
             Py_DECREF(io);
             PyErr_SetString(PyExc_TypeError,
-                            "nb_int should return int object");
+                        "__int__ method should return an integer");
             return (unsigned long)-1;
         }
     }
@@ -324,7 +326,7 @@ PyInt_AsUnsignedLongLongMask(register PyObject *op)
         {
             Py_DECREF(io);
             PyErr_SetString(PyExc_TypeError,
-                            "nb_int should return int object");
+                        "__int__ method should return an integer");
             return (unsigned PY_LONG_LONG)-1;
         }
     }
@@ -434,12 +436,6 @@ int_print(PyIntObject *v, FILE *fp, int flags)
     return 0;
 }
 
-static PyObject *
-int_repr(PyIntObject *v)
-{
-    return _PyInt_Format(v, 10, 0);
-}
-
 static int
 int_compare(PyIntObject *v, PyIntObject *w)
 {
@@ -649,16 +645,35 @@ int_classic_div(PyIntObject *x, PyIntObject *y)
 }
 
 static PyObject *
-int_true_divide(PyObject *v, PyObject *w)
+int_true_divide(PyIntObject *x, PyIntObject *y)
 {
+    long xi, yi;
     /* If they aren't both ints, give someone else a chance.  In
        particular, this lets int/long get handled by longs, which
        underflows to 0 gracefully if the long is too big to convert
        to float. */
-    if (PyInt_Check(v) && PyInt_Check(w))
-        return PyFloat_Type.tp_as_number->nb_true_divide(v, w);
-    Py_INCREF(Py_NotImplemented);
-    return Py_NotImplemented;
+    CONVERT_TO_LONG(x, xi);
+    CONVERT_TO_LONG(y, yi);
+    if (yi == 0) {
+        PyErr_SetString(PyExc_ZeroDivisionError,
+                        "division by zero");
+        return NULL;
+    }
+    if (xi == 0)
+        return PyFloat_FromDouble(yi < 0 ? -0.0 : 0.0);
+
+#define WIDTH_OF_ULONG (CHAR_BIT*SIZEOF_LONG)
+#if DBL_MANT_DIG < WIDTH_OF_ULONG
+    if ((xi >= 0 ? 0UL + xi : 0UL - xi) >> DBL_MANT_DIG ||
+        (yi >= 0 ? 0UL + yi : 0UL - yi) >> DBL_MANT_DIG)
+        /* Large x or y.  Use long integer arithmetic. */
+        return PyLong_Type.tp_as_number->nb_true_divide(
+            (PyObject *)x, (PyObject *)y);
+    else
+#endif
+        /* Both ints can be exactly represented as doubles.  Do a
+           floating-point division. */
+        return PyFloat_FromDouble((double)xi / (double)yi);
 }
 
 static PyObject *
@@ -736,7 +751,13 @@ int_pow(PyIntObject *v, PyIntObject *w, PyIntObject *z)
     while (iw > 0) {
         prev = ix;              /* Save value for overflow check */
         if (iw & 1) {
-            ix = ix*temp;
+            /*
+             * The (unsigned long) cast below ensures that the multiplication
+             * is interpreted as an unsigned operation rather than a signed one
+             * (C99 6.3.1.8p1), thus avoiding the perils of undefined behaviour
+             * from signed arithmetic overflow (C99 6.5p5).  See issue #12973.
+             */
+            ix = (unsigned long)ix * temp;
             if (temp == 0)
                 break; /* Avoid ix / 0 */
             if (ix / temp != prev) {
@@ -749,7 +770,7 @@ int_pow(PyIntObject *v, PyIntObject *w, PyIntObject *z)
         iw >>= 1;               /* Shift exponent down by 1 bit */
         if (iw==0) break;
         prev = temp;
-        temp *= temp;           /* Square the value of temp */
+        temp = (unsigned long)temp * temp;  /* Square the value of temp */
         if (prev != 0 && temp / prev != prev) {
             return PyLong_Type.tp_as_number->nb_power(
                 (PyObject *)v, (PyObject *)w, (PyObject *)z);
@@ -939,12 +960,78 @@ int_long(PyIntObject *v)
     return PyLong_FromLong((v -> ob_ival));
 }
 
+static const unsigned char BitLengthTable[32] = {
+    0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
+    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5
+};
+
+static int
+bits_in_ulong(unsigned long d)
+{
+    int d_bits = 0;
+    while (d >= 32) {
+        d_bits += 6;
+        d >>= 6;
+    }
+    d_bits += (int)BitLengthTable[d];
+    return d_bits;
+}
+
+#if 8*SIZEOF_LONG-1 <= DBL_MANT_DIG
+/* Every Python int can be exactly represented as a float. */
+
 static PyObject *
 int_float(PyIntObject *v)
 {
     return PyFloat_FromDouble((double)(v -> ob_ival));
 }
 
+#else
+/* Here not all Python ints are exactly representable as floats, so we may
+   have to round.  We do this manually, since the C standards don't specify
+   whether converting an integer to a float rounds up or down */
+
+static PyObject *
+int_float(PyIntObject *v)
+{
+    unsigned long abs_ival, lsb;
+    int round_up;
+
+    if (v->ob_ival < 0)
+        abs_ival = 0U-(unsigned long)v->ob_ival;
+    else
+        abs_ival = (unsigned long)v->ob_ival;
+    if (abs_ival < (1L << DBL_MANT_DIG))
+        /* small integer;  no need to round */
+        return PyFloat_FromDouble((double)v->ob_ival);
+
+    /* Round abs_ival to MANT_DIG significant bits, using the
+       round-half-to-even rule.  abs_ival & lsb picks out the 'rounding'
+       bit: the first bit after the most significant MANT_DIG bits of
+       abs_ival.  We round up if this bit is set, provided that either:
+
+         (1) abs_ival isn't exactly halfway between two floats, in which
+         case at least one of the bits following the rounding bit must be
+         set; i.e., abs_ival & lsb-1 != 0, or:
+
+         (2) the resulting rounded value has least significant bit 0; or
+         in other words the bit above the rounding bit is set (this is the
+         'to-even' bit of round-half-to-even); i.e., abs_ival & 2*lsb != 0
+
+       The condition "(1) or (2)" equates to abs_ival & 3*lsb-1 != 0. */
+
+    lsb = 1L << (bits_in_ulong(abs_ival)-DBL_MANT_DIG-1);
+    round_up = (abs_ival & lsb) && (abs_ival & (3*lsb-1));
+    abs_ival &= -2*lsb;
+    if (round_up)
+        abs_ival += 2*lsb;
+    return PyFloat_FromDouble(v->ob_ival < 0 ?
+                              -(double)abs_ival :
+                  (double)abs_ival);
+}
+
+#endif
+
 static PyObject *
 int_oct(PyIntObject *v)
 {
@@ -1048,8 +1135,33 @@ int_getnewargs(PyIntObject *v)
 }
 
 static PyObject *
-int_getN(PyIntObject *v, void *context) {
-    return PyInt_FromLong((Py_intptr_t)context);
+int_get0(PyIntObject *v, void *context) {
+    return PyInt_FromLong(0L);
+}
+
+static PyObject *
+int_get1(PyIntObject *v, void *context) {
+    return PyInt_FromLong(1L);
+}
+
+/* Convert an integer to a decimal string.  On many platforms, this
+   will be significantly faster than the general arbitrary-base
+   conversion machinery in _PyInt_Format, thanks to optimization
+   opportunities offered by division by a compile-time constant. */
+static PyObject *
+int_to_decimal_string(PyIntObject *v) {
+    char buf[sizeof(long)*CHAR_BIT/3+6], *p, *bufend;
+    long n = v->ob_ival;
+    unsigned long absn;
+    p = bufend = buf + sizeof(buf);
+    absn = n < 0 ? 0UL - n : n;
+    do {
+        *--p = '0' + (char)(absn % 10);
+        absn /= 10;
+    } while (absn);
+    if (n < 0)
+        *--p = '-';
+    return PyString_FromStringAndSize(p, bufend - p);
 }
 
 /* Convert an integer to the given base.  Returns a string.
@@ -1076,6 +1188,10 @@ _PyInt_Format(PyIntObject *v, int base, int newstyle)
 
     assert(base >= 2 && base <= 36);
 
+    /* Special case base 10, for speed */
+    if (base == 10)
+        return int_to_decimal_string(v);
+
     do {
         /* I'd use i_divmod, except it doesn't produce the results
            I want when n is negative.  So just duplicate the salient
@@ -1108,7 +1224,7 @@ _PyInt_Format(PyIntObject *v, int base, int newstyle)
         *--p = 'x';
         *--p = '0';
     }
-    else if (base != 10) {
+    else {
         *--p = '#';
         *--p = '0' + base%10;
         if (base > 10)
@@ -1150,6 +1266,29 @@ int__format__(PyObject *self, PyObject *args)
     return NULL;
 }
 
+static PyObject *
+int_bit_length(PyIntObject *v)
+{
+    unsigned long n;
+
+    if (v->ob_ival < 0)
+        /* avoid undefined behaviour when v->ob_ival == -LONG_MAX-1 */
+        n = 0U-(unsigned long)v->ob_ival;
+    else
+        n = (unsigned long)v->ob_ival;
+
+    return PyInt_FromLong(bits_in_ulong(n));
+}
+
+PyDoc_STRVAR(int_bit_length_doc,
+"int.bit_length() -> int\n\
+\n\
+Number of bits necessary to represent self in binary.\n\
+>>> bin(37)\n\
+'0b100101'\n\
+>>> (37).bit_length()\n\
+6");
+
 #if 0
 static PyObject *
 int_is_finite(PyObject *v)
@@ -1161,6 +1300,8 @@ int_is_finite(PyObject *v)
 static PyMethodDef int_methods[] = {
     {"conjugate",       (PyCFunction)int_int,   METH_NOARGS,
      "Returns self, the complex conjugate of any int."},
+    {"bit_length", (PyCFunction)int_bit_length, METH_NOARGS,
+     int_bit_length_doc},
 #if 0
     {"is_finite",       (PyCFunction)int_is_finite,     METH_NOARGS,
      "Returns always True."},
@@ -1178,17 +1319,17 @@ static PyGetSetDef int_getset[] = {
      "the real part of a complex number",
      NULL},
     {"imag",
-     (getter)int_getN, (setter)NULL,
+     (getter)int_get0, (setter)NULL,
      "the imaginary part of a complex number",
-     (void*)0},
+     NULL},
     {"numerator",
      (getter)int_int, (setter)NULL,
      "the numerator of a rational number in lowest terms",
      NULL},
     {"denominator",
-     (getter)int_getN, (setter)NULL,
+     (getter)int_get1, (setter)NULL,
      "the denominator of a rational number in lowest terms",
-     (void*)1},
+     NULL},
     {NULL}  /* Sentinel */
 };
 
@@ -1239,7 +1380,7 @@ static PyNumberMethods int_as_number = {
     0,                          /*nb_inplace_xor*/
     0,                          /*nb_inplace_or*/
     (binaryfunc)int_div,        /* nb_floor_divide */
-    int_true_divide,            /* nb_true_divide */
+    (binaryfunc)int_true_divide, /* nb_true_divide */
     0,                          /* nb_inplace_floor_divide */
     0,                          /* nb_inplace_true_divide */
     (unaryfunc)int_int,         /* nb_index */
@@ -1255,13 +1396,13 @@ PyTypeObject PyInt_Type = {
     0,                                          /* tp_getattr */
     0,                                          /* tp_setattr */
     (cmpfunc)int_compare,                       /* tp_compare */
-    (reprfunc)int_repr,                         /* tp_repr */
+    (reprfunc)int_to_decimal_string,            /* tp_repr */
     &int_as_number,                             /* tp_as_number */
     0,                                          /* tp_as_sequence */
     0,                                          /* tp_as_mapping */
     (hashfunc)int_hash,                         /* tp_hash */
     0,                                          /* tp_call */
-    (reprfunc)int_repr,                         /* tp_str */
+    (reprfunc)int_to_decimal_string,            /* tp_str */
     PyObject_GenericGetAttr,                    /* tp_getattro */
     0,                                          /* tp_setattro */
     0,                                          /* tp_as_buffer */
